"""
Convert a "memorious collection" (the output format of the store->directory
stage) into a leakrfc dataset

memorious format:
    ./data/store/test_dataset/
        ./<sha1>.data.pdf|doc|...  # actual file
        ./<sha1>.json              # metadata file
"""

from pathlib import Path
from typing import Any, Callable
from urllib.parse import unquote, urlparse

from anystore import anycache
from anystore.store import get_store
from anystore.types import StrGenerator, Uri
from anystore.util import make_data_checksum
from anystore.worker import WorkerStatus

from leakrfc.archive import DatasetArchive
from leakrfc.archive.cache import get_cache
from leakrfc.logging import get_logger
from leakrfc.model import OriginalFile
from leakrfc.util import render
from leakrfc.worker import DatasetWorker

log = get_logger(__name__)


def make_cache_key(self: "MemoriousWorker", key: str) -> str | None:
    if not self.use_cache:
        return
    host = urlparse(self.memorious.uri).netloc
    if host is None:
        host = make_data_checksum(str(self.memorious.uri))
    return f"memorious/sync/{host}/{self.dataset.name}/{key}"


class MemoriousStatus(WorkerStatus):
    added: int = 0
    skipped: int = 0
    not_found: int = 0


class MemoriousWorker(DatasetWorker):
    def __init__(
        self, uri: Uri, key_func: Callable | None = None, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.memorious = get_store(uri, serialization_mode="raw")
        self.key_func = key_func or get_file_key
        self.status_model = MemoriousStatus

    def get_tasks(self) -> StrGenerator:
        yield from self.memorious.iterate_keys(glob="*.json")

    def handle_task(self, task: str) -> None:
        file = self.load_memorious(task)
        if file is not None:
            if not self.dataset.exists(file.key):
                self.dataset.archive_file(
                    file.extra.pop("_file_name"),
                    store=self.memorious,
                    file=file,
                )
                self.count(added=1)
            else:
                self.log_info(
                    f"Skipping already existing `{file.key}` ...",
                    store=self.memorious.uri,
                )
                self.count(skipped=1)

    @anycache(store=get_cache(), key_func=make_cache_key, model=OriginalFile)
    def load_memorious(self, key: str) -> OriginalFile | None:
        data = self.memorious.get(key, serialization_mode="json")
        content_hash = data.pop("content_hash", None)
        if content_hash is None:
            log.warning(f"No content hash for `{key}`", store=self.memorious.uri)
            self.count(not_found=1)
        elif data.get("_file_name") is None:
            log.warning(f"No original file for `{key}`", store=self.memorious.uri)
        else:
            key = self.key_func(data)
            info = self.memorious.info(data["_file_name"])
            return OriginalFile(
                key=key.strip("/"),
                name=Path(key).name,
                size=info.size,
                content_hash=content_hash,
                store=str(self.memorious.uri),
                dataset=self.dataset.name,
                extra=data,
            )

    def done(self) -> None:
        self.log_info(f"Done memorious import from `{self.memorious.uri}`")


def import_memorious(
    dataset: DatasetArchive,
    uri: Uri,
    key_func: Callable | None = None,
    use_cache: bool | None = True,
) -> MemoriousStatus:
    worker = MemoriousWorker(uri, key_func, dataset=dataset, use_cache=use_cache)
    worker.log_info(f"Starting memorious import from `{worker.memorious.uri}` ...")
    return worker.run()


def get_file_key(data: dict[str, Any]) -> str:
    return unquote(urlparse(data["url"]).path).strip("/")


def get_file_name(data: dict[str, Any]) -> str:
    return unquote(Path(urlparse(data["url"]).path).name)


def get_file_name_templ_func(tmpl: str) -> Callable:
    def _func(data: dict[str, Any]) -> str:
        return render(tmpl, data)

    return _func


def get_file_name_strip_func(strip_prefix: str) -> Callable:
    def _func(data: dict[str, Any]) -> str:
        return get_file_key(data)[len(strip_prefix) :].strip("/")

    return _func



import os
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Generator

import fsspec
import yaml
from anystore import get_store as _get_store
from anystore.store import Store, ZipStore

from leakrfc.logging import get_logger
from leakrfc.model import ArchiveModel

if TYPE_CHECKING:
    from leakrfc.archive.dataset import DatasetArchive


log = get_logger(__name__)


OPTS = {"serialization_mode": "raw"}


def get_store(**kwargs) -> Store | ZipStore:
    uri = kwargs.get("uri")
    if uri and os.path.splitext(uri)[1] == ".leakrfc":
        return ZipStore(**kwargs)
    return _get_store(**kwargs)


class BaseArchive(ArchiveModel):
    @cached_property
    def _storage(self) -> Store:
        if self.storage is not None:
            config = {**self.storage.model_dump(), **OPTS}
            return get_store(**config)
        return get_store(uri=self.uri, **OPTS)

    @cached_property
    def is_zip(self) -> bool:
        return isinstance(self._storage, ZipStore)

    def _make_path(self, *parts: str) -> str:
        return "/".join([p.strip("/") for p in parts if p.strip("/")])


class Archive(BaseArchive):
    """
    Leakrfc archive that holds one or more datasets as subdirs
    """

    def get_dataset(self, dataset: str) -> "DatasetArchive":
        from leakrfc.archive.dataset import DatasetArchive

        config_uri = f"{dataset}/{self.metadata_prefix}/config.yml"
        config = {}
        if self._storage.exists(config_uri):
            config = self._storage.get(config_uri, deserialization_func=yaml.safe_load)
        if "storage" not in config:
            storage = self._storage.model_dump()
            if not self.is_zip:
                storage["uri"] = self._make_path(self._storage.uri, dataset)
            config.update(storage=storage)
        config["name"] = dataset
        return DatasetArchive(**config)

    def get_datasets(self) -> Generator["DatasetArchive", None, None]:
        fs, _ = fsspec.url_to_fs(str(self._storage.uri))
        for child in fs.ls(self._storage.uri):
            dataset = Path(child).name
            if self._storage.exists(f"{dataset}/{self.metadata_prefix}"):
                yield self.get_dataset(dataset)



"""
Make or update a leakrfc dataset and check integrity
"""

from datetime import datetime
from typing import Generator, Literal, TypeAlias

from anystore.decorators import anycache
from anystore.exceptions import DoesNotExist
from anystore.worker import WorkerStatus

from leakrfc.archive.cache import get_cache
from leakrfc.archive.dataset import DatasetArchive
from leakrfc.worker import DatasetWorker


class MakeStatus(WorkerStatus):
    files_total: int = 0
    metadata_total: int = 0
    files_added: int = 0
    files_updated: int = 0
    files_deleted: int = 0
    integrity_errors: int = 0


ACTION_SOURCE = "source"
ACTION_INFO = "info"
Action: TypeAlias = Literal["info", "source"]
Task: TypeAlias = tuple[str, Action]


def make_cache_key(self: "LeakrfcWorker", task: Task) -> str | None:
    if self.use_cache:
        key, action = task
        return f"{self.dataset.name}/make/{action}/{key}"


def make_cache_key_integrity(self: "LeakrfcWorker", key: str) -> str | None:
    if self.use_cache:
        return f"{self.dataset.name}/make/integrity/{key}"


class LeakrfcWorker(DatasetWorker):
    def __init__(
        self,
        check_integrity: bool | None = True,
        cleanup: bool | None = True,
        *args,
        **kwargs,
    ) -> None:
        kwargs["status_model"] = kwargs.get("status_model", MakeStatus)
        super().__init__(*args, **kwargs)
        self.check_integrity = check_integrity
        self.cleanup = cleanup

    def get_tasks(self) -> Generator[Task, None, None]:
        self.log_info("Checking source files ...")
        for key in self.dataset.iter_keys():
            self.count(files_total=1)
            yield key, ACTION_SOURCE
        self.log_info("Checking existing files ...")
        for file in super().get_tasks():
            self.count(metadata_total=1)
            yield file.key, ACTION_INFO

    @anycache(store=get_cache(), key_func=make_cache_key)
    def handle_task(self, task: Task) -> str:
        key, action = task
        now = datetime.now().isoformat()
        if action == ACTION_SOURCE:
            self.log_info(f"Checking `{key}` ...", action=action)
            if not self.dataset.exists(key):
                self.dataset.archive_file(key, self.dataset._storage)
                self.count(files_added=1)
            self._ensure_integrity(key)
        elif action == ACTION_INFO:
            self.log_info(f"Checking `{key}` metadata ...", action=action)
            self._ensure_integrity(key)
        return now

    @anycache(store=get_cache(), key_func=make_cache_key_integrity)
    def _ensure_integrity(self, key: str) -> None:
        if self.check_integrity:
            self.count(files_checked=1)
            self.log_info(f"Testing checksum for `{key}` ...")
            try:
                content_hash = self.dataset.make_checksum(key)
                file = self.dataset.lookup_file(key)
                if content_hash != file.content_hash:
                    self.log_error(
                        f"Checksum mismatch for `{key}`: `{content_hash}`",
                        file=file,
                    )
                    self.count(integrity_errors=1)
                    if self.cleanup:
                        self.log_info(f"Fixing checksum for `{key}` ...")
                        file.content_hash = content_hash
                        self.dataset._put_file_info(file)
            except DoesNotExist:
                self.log_error(f"Source file `{key}` does not exist")
                self.count(files_deleted=1)
                if self.cleanup:
                    self.log_info(f"Deleting metadata for `{key}` ...")
                    self.dataset.delete_file(key)


def make_dataset(
    dataset: DatasetArchive,
    use_cache: bool | None = True,
    check_integrity: bool | None = True,
    cleanup: bool | None = True,
) -> MakeStatus:
    worker = LeakrfcWorker(check_integrity, cleanup, dataset, use_cache=use_cache)
    return worker.run()



from functools import cache

from anystore import get_store
from anystore.store import BaseStore

from leakrfc.logging import get_logger

log = get_logger(__name__)


@cache
def get_cache() -> BaseStore:
    from leakrfc.settings import ArchiveSettings

    settings = ArchiveSettings()
    if settings.cache is not None:
        return get_store(**settings.cache.model_dump())
    log.warning(
        "Using in-memory cache. This is not for production use! "
        "Configure via env `LEAKRFC_CACHE__*`"
    )
    return get_store("memory:///")



import os
from functools import cache
from pathlib import Path
from typing import Any

from alephclient.api import AlephAPI
from alephclient.settings import API_KEY, HOST

from leakrfc.logging import get_logger

log = get_logger(__name__)


class AlephException(BaseException):
    pass


@cache
def get_api(host: str | None = None, api_key: str | None = None) -> AlephAPI:
    return AlephAPI(host=host or HOST, api_key=api_key or API_KEY)


@cache
def get_host(api: AlephAPI | None = None) -> str:
    api = api or get_api()
    return api.base_url[:-7]


@cache
def get_foreign_id(collection_id: str, api: AlephAPI | None = None) -> str:
    api = api or get_api()
    res = api.get_collection(collection_id)
    if res is None:
        raise AlephException(
            "Collection with collection_id `%s` not found or not accessible."
            % collection_id
        )
    return res["foreign_id"]


@cache
def get_collection_id(foreign_id: str, api: AlephAPI | None = None) -> str:
    api = api or get_api()
    res = api.get_collection_by_foreign_id(foreign_id)
    if res is None:
        raise AlephException(
            "Collection with foreign_id `%s` not found or not accessible." % foreign_id
        )
    return res["id"]


@cache
def get_or_create_collection_id(foreign_id: str, api: AlephAPI | None = None) -> str:
    api = api or get_api()
    res = api.load_collection_by_foreign_id(foreign_id)
    return res["id"]


@cache
def make_folders(path: str, collection_id: str, parent: str | None = None) -> str:
    api = get_api()
    log.info(f"Creating folder: `{path}`", host=get_host(api))
    folder = Path(path)
    foreign_id = "/".join(folder.parts)
    if len(folder.parts) > 1:
        parent = make_folders(os.path.join(*folder.parts[:-1]), collection_id, parent)
    metadata: dict[str, Any] = {"file_name": folder.name, "foreign_id": foreign_id}
    if parent is not None:
        metadata["parent"] = {"id": parent}
    res = api.ingest_upload(collection_id, metadata=metadata)
    return res["id"]



from datetime import datetime
from typing import Any, ClassVar, Generator, Self, TypeAlias

from anystore.mixins import BaseModel
from anystore.model import StoreModel
from anystore.store import get_store_for_uri
from anystore.store.base import Stats
from anystore.types import Uri
from anystore.util import make_data_checksum
from ftmq.util import make_proxy
from nomenklatura.dataset import DefaultDataset
from nomenklatura.entity import CE
from pantomime import DEFAULT
from pydantic import field_validator, model_validator

from leakrfc.util import guess_mimetype

ORIGIN_ORIGINAL = "original"
ORIGIN_EXTRACTED = "extracted"
ORIGIN_CONVERTED = "converted"


class ArchiveModel(BaseModel):
    uri: str | None = None
    metadata_prefix: str = ".leakrfc"
    public_url: str | None = None
    checksum_algorithm: str = "sha1"
    storage: StoreModel | None = None


class File(Stats):
    dataset: str
    content_hash: str
    mimetype: str | None = None
    processed: datetime | None = None
    extra: dict[str, Any] = {}

    def model_dump(self, *args, **kwargs) -> dict[str, Any]:
        data = super().model_dump(*args, **kwargs)
        if hasattr(self, "origin"):
            data["origin"] = self.origin
        return data

    def to_proxy(self) -> CE:
        proxy = make_proxy({"id": self.id, "schema": "Document"}, dataset=self.dataset)
        proxy.add("contentHash", self.content_hash)
        proxy.add("fileName", self.name)
        proxy.add("fileSize", self.size)
        proxy.add("mimeType", self.mimetype)
        return proxy

    @property
    def id(self) -> str:
        return (
            f"{self.dataset}-file-{make_data_checksum((self.key, self.content_hash))}"
        )

    @classmethod
    def from_info(cls, info: Stats, dataset: str, **data) -> Self:
        data["dataset"] = dataset
        return cls(**{**info.model_dump(), **data})

    @classmethod
    def from_uri(cls, uri: Uri, dataset: str | None = None, **data) -> Self:
        if dataset is None:
            dataset = DefaultDataset.name
        store, uri = get_store_for_uri(uri)
        return cls.from_info(store.info(uri), dataset, **data)

    @classmethod
    @field_validator("mimetype")
    def normalize_mimetype(cls, v: Any) -> str | None:
        return guess_mimetype(v)

    @model_validator(mode="after")
    def assign_mimetype(self):
        if self.mimetype in (None, DEFAULT):
            self.mimetype = guess_mimetype(self.name) or DEFAULT
        return self


class OriginalFile(File):
    origin: ClassVar = ORIGIN_ORIGINAL


class ExtractedFile(File):
    origin: ClassVar = ORIGIN_EXTRACTED
    parent: str
    root: str


class ConvertedFile(File):
    origin: ClassVar = ORIGIN_CONVERTED
    root: str


OriginalFiles: TypeAlias = Generator[OriginalFile, None, None]
Files: TypeAlias = Generator[File, None, None]



# from anystore.io import smart_open
# from anystore.mixins import orjson
# from anystore.store import ZipStore
# from anystore.types import Uri

# from leakrfc.archive import Archive
# from leakrfc.logging import get_logger
# from leakrfc.model import File
# from leakrfc.util import make_ch_key

# log = get_logger(__name__)


# def import_dataset(uri: Uri, archive: Archive) -> None:
#     log.info(f"Importing dataset `{uri}`...", archive=archive.file_store.uri)
#     in_file = ZipStore(uri=uri, prefix=dataset.name)
#     proxies = b""
#     for info in dataset.iter_files():
#         proxies += orjson.dumps(
#             info.to_proxy().to_dict(), option=orjson.OPT_APPEND_NEWLINE
#         )
#         meta_key = f".leakrfc/meta/{make_ch_key(info.content_hash)}"
#         out.put(meta_key, info, model=File)
#         log.info(
#             f"Adding file `{info.name} ({info.content_hash})` metadata ...",
#             dataset=dataset.name,
#             store=dataset.file_store.uri,
#         )
#         with out.open(info.key, mode="wb") as o:
#             o.write(i.read())
#     out.put(".leakrfc/entities/entities.ftm.json", proxies)



import os
import tempfile
from os import PathLike
from pathlib import Path
from typing import Any

import shortuuid
from anystore import get_store as get_file_store
from followthemoney import model
from followthemoney.proxy import EntityProxy
from ftmq.store import get_store as get_ftm_store
from ftmq.util import make_proxy
from ingestors.manager import Manager

from leakrfc.archive.dataset import DatasetArchive
from leakrfc.logging import get_logger
from leakrfc.worker import DatasetWorker

log = get_logger(__name__)


class IngestManager(Manager):
    def __init__(
        self,
        dataset: DatasetArchive,
        worker: DatasetWorker,
        context: dict[str, Any] | None = None,
    ):
        store = get_ftm_store()
        self.writer = store.writer()
        self.dataset = dataset
        self.context = context
        self.work_path = Path(tempfile.mkdtemp(prefix="leakrfc-ingestor-"))
        self.work_store = get_file_store(self.work_path)
        self.emitted = set()
        self.worker = worker

    @property
    def archive(self):
        raise NotImplementedError("Accessing archive not implemented")

    def make_entity(self, schema: str, parent=None) -> EntityProxy:
        entity = model.make_entity(schema)
        self.make_child(parent, entity)
        return entity

    def emit_entity(self, entity: EntityProxy, *args, **kwargs):
        entity = make_proxy(entity.to_dict(), dataset=self.dataset.name)
        self.writer.add_entity(entity)
        self.emitted.add(entity.id)

    def queue_entity(self, entity: EntityProxy):
        self.worker.queue.put(entity.to_full_dict())

    def store(self, file_path: PathLike, *args, **kwargs) -> str:
        file = self.dataset.archive_file(file_path)
        return file.content_hash

    def load(self, content_hash: str, file_name: str | None = None) -> Path:
        file_name = file_name or shortuuid.uuid()
        path = os.path.join(content_hash, file_name)
        with self.dataset.open_file(content_hash) as i:
            with self.work_store.open(path, mode="wb") as o:
                o.write(i.read())
        return (self.work_path / path).resolve().absolute()


class IngestWorker(DatasetWorker):
    def get_tasks(self) -> Any:
        for file in super().get_tasks():
            yield file.to_proxy().to_full_dict()

    def handle_task(self, task: dict[str, Any]) -> None:
        entity = model.get_proxy(task)
        manager = IngestManager(self.dataset, self)
        log.debug("Ingest: %r", entity)
        try:
            manager.ingest_entity(entity)
        finally:
            manager.close()
        log.debug("Created %d entities", len(manager.emitted))


def ingest_dataset(dataset: DatasetArchive) -> None:
    worker = IngestWorker(dataset)
    worker.log_info(f"Ingesting dataset `{dataset.name}` ...")
    worker.run()



import os
from functools import cache

from anystore.model import StoreModel
from anystore.types import Uri
from anystore.util import ensure_uri

from leakrfc.archive.base import Archive
from leakrfc.archive.dataset import DatasetArchive, ReadOnlyDatasetArchive


def configure_archive(**kwargs) -> Archive:
    """change config during tests runtime"""
    from leakrfc.settings import ArchiveSettings

    settings = ArchiveSettings()
    if settings.uri is not None:
        return get_archive(settings.uri)
    return Archive(**{**settings.model_dump(), **kwargs})


@cache
def get_archive(uri: Uri | None = None, **kwargs) -> Archive:
    if uri is not None:
        uri = ensure_uri(uri)
        ext = os.path.splitext(uri)[1]
        if ext in (".yml", ".yaml"):
            return Archive._from_uri(uri, **kwargs)
        else:
            return Archive(storage=StoreModel(uri=uri, **kwargs))
    return configure_archive()


@cache
def get_dataset(
    dataset: str, uri: Uri | None = None, **kwargs
) -> DatasetArchive | ReadOnlyDatasetArchive:
    if uri is not None:
        return DatasetArchive(name=dataset, uri=str(uri), **kwargs)
    archive = get_archive(**kwargs)
    return archive.get_dataset(dataset)


archive = get_archive()



"""
Crawl document collections from public accessible archives (or local folders)
"""

from typing import Generator

from anystore import get_store
from anystore.store import BaseStore
from anystore.types import Uri

from leakrfc.archive import DatasetArchive
from leakrfc.logging import get_logger
from leakrfc.worker import DatasetWorker

log = get_logger(__name__)


class CrawlWorker(DatasetWorker):
    def __init__(self, remote: BaseStore, **kwargs) -> None:
        super().__init__(**kwargs)
        self.remote = remote

    def get_tasks(self) -> Generator[str, None, None]:
        self.log_info(f"Crawling `{self.remote.uri}` ...")
        yield from self.remote.iterate_keys()

    def handle_task(self, task: str) -> None:
        self.log_info(f"Crawling `{task}` ...", remote=self.remote.uri)
        self.dataset.archive_file(task, self.remote)

    def done(self) -> None:
        self.log_info(f"Crawling `{self.remote.uri}`: Done.")


def crawl(uri: Uri, storage: DatasetArchive) -> None:
    remote_store = get_store(uri=uri, serialization_mode="raw")
    worker = CrawlWorker(remote_store, dataset=storage)
    worker.run()



from typing import Optional

from anystore.io import smart_read
from anystore.model import StoreModel
from pydantic import BaseModel, HttpUrl
from pydantic_settings import BaseSettings, SettingsConfigDict

from leakrfc.model import ArchiveModel


class Settings(BaseSettings):
    # model_config = SettingsConfigDict(cli_parse_args=True)

    debug: bool = False
    log_json: bool = False
    log_level: str = "INFO"

    aleph_secret_key: str = (
        "aleph-scrt"  # to act as a servicelayer replacement to the ui
    )


class ArchiveSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="leakrfc_",
        env_nested_delimiter="__",
        nested_model_default_partial_update=True,
        # cli_parse_args=True
    )

    uri: str | None = None
    archive: ArchiveModel | None = None
    cache: StoreModel | None = None


class ApiContactSettings(BaseModel):
    name: str | None
    url: str | None
    email: str | None


class ApiSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="leakrfc_api_",
        env_nested_delimiter="__",
        nested_model_default_partial_update=True,
    )

    secret_key: str = "change-for-production"
    access_token_expire: int = 5  # minutes
    access_token_algorithm: str = "HS256"

    title: str = "LeakRFC Api"
    description: str = smart_read("./README.md", mode="r")
    contact: ApiContactSettings | None = None

    allowed_origin: Optional[HttpUrl] = "http://localhost:3000"



from datetime import datetime
from typing import Any, BinaryIO, ClassVar, Iterable

import orjson
from anystore.store.base import BaseStore
from anystore.store.virtual import get_virtual
from anystore.types import BytesGenerator, StrGenerator
from anystore.util import DEFAULT_HASH_ALGORITHM
from banal import ensure_dict
from nomenklatura.entity import CE

from leakrfc.archive.base import BaseArchive, get_store
from leakrfc.logging import get_logger
from leakrfc.model import OriginalFile, OriginalFiles
from leakrfc.util import make_ch_key

log = get_logger(__name__)

INFO_PREFIX = "info"
INFO = "info.json"
STORE_PREFIX = "store"
KEY = "key"
TXT = "txt"
ENTITIES_PREFIX = "entities"
ENTITIES = "entities.ftm.json"


class ReadOnlyDatasetArchive(BaseArchive):
    readonly: ClassVar = True
    name: str

    def exists(self, key: str) -> bool:
        path = self._get_file_info_path(key)
        return self._storage.exists(path)

    def exists_hash(self, content_hash: str) -> bool:
        lookup_key = self._make_path(
            self.metadata_prefix,
            STORE_PREFIX,
            make_ch_key(content_hash),
            KEY,
        )
        key = self._storage.get(
            lookup_key, raise_on_nonexist=False, serialization_mode="auto"
        )
        if key is None:
            return False
        return self.exists(key)

    def make_checksum(
        self, key: str, algorithm: str | None = DEFAULT_HASH_ALGORITHM
    ) -> str:
        return self._storage.checksum(self._make_path(key), algorithm)

    def lookup_file(self, key: str) -> OriginalFile:
        path = self._get_file_info_path(key)
        return self._storage.get(path, model=OriginalFile)

    def lookup_file_by_hash(self, content_hash: str) -> OriginalFile:
        key_path = self._make_path(
            self.metadata_prefix, STORE_PREFIX, make_ch_key(content_hash), KEY
        )
        key = self._storage.get(key_path, serialization_mode="auto")
        return self.lookup_file(key)

    def stream_file(self, file: OriginalFile) -> BytesGenerator:
        yield from self._storage.stream(self._make_path(file.key))

    def open_file(self, file: OriginalFile) -> BinaryIO:
        return self._storage.open(self._make_path(file.key))

    def iter_files(self) -> OriginalFiles:
        prefix = self._make_path(self.metadata_prefix, INFO_PREFIX)
        for key in self._storage.iterate_keys(prefix=prefix):
            yield self._storage.get(key, model=OriginalFile)

    def iter_keys(self) -> StrGenerator:
        for key in self._storage.iterate_keys(
            prefix=self._make_path(),
            exclude_prefix=self._make_path(self.metadata_prefix),
        ):
            if self.is_zip:
                key = key[len(self.name) + 1 :]
            yield key

    def _get_file_info_path(self, key) -> str:
        return self._make_path(self.metadata_prefix, INFO_PREFIX, key, INFO)

    def _get_file_store_path(self, key: str, *parts: str) -> str:
        file = self.lookup_file(key)
        return self._make_path(
            self.metadata_prefix, STORE_PREFIX, make_ch_key(file.content_hash), *parts
        )

    def _get_entities_path(
        self, now: bool = False, suffix: str | None = None, with_prefix: bool = False
    ) -> str:
        path = ENTITIES
        if now:
            suffix = datetime.now().isoformat()
        if suffix:
            path += f".{suffix}"
        path = self._make_path(ENTITIES_PREFIX, path)
        if with_prefix:
            path = self._make_path(self.metadata_prefix, path)
        return path

    def _make_path(self, *parts: str) -> str:
        if self.is_zip:
            parts = tuple([self.name, *parts])
        return super()._make_path(*parts)


class DatasetArchive(ReadOnlyDatasetArchive):
    readonly: ClassVar = False

    def archive_file(
        self,
        uri: str,
        store: BaseStore | None = None,
        file: OriginalFile | None = None,
        data: dict[str, Any] | None = None,
    ) -> OriginalFile:
        """Add the given file to the archive. This doesn't check for existing
        files or if the given `file.content_hash` is correct. This should be
        handled in higher logic, as seen in `leakrfc.make.make_dataset`."""

        if store is None:
            store_uri, uri = uri.rsplit("/", 1)
            store = get_store(uri=store_uri)
        log.debug(
            "Adding file ...",
            uri=uri,
            dataset=self.name,
            from_store=store.uri,
            to_store=self._storage.uri,
        )

        with get_virtual(prefix=f"leakrfc-{self.name}-") as tmp:
            tmp_path = tmp.download(uri, store)
            if file is not None:
                content_hash = file.content_hash
            else:
                content_hash = tmp.store.checksum(tmp_path, self.checksum_algorithm)

            if file is None:
                file = OriginalFile.from_info(
                    store.info(uri), dataset=self.name, content_hash=content_hash
                )
            file_data = file.model_dump()
            file_data.update(
                {
                    **ensure_dict(data),
                    "content_hash": content_hash,
                    "dataset": self.name,
                    "store": self._storage.uri,
                }
            )
            file = OriginalFile(**file_data)
            # store actual file
            file_path = self._make_path(file.key)
            with tmp.store.open(tmp_path) as i:
                with self._storage.open(file_path, mode="wb") as o:
                    o.write(i.read())
            # store metadata
            self._put_file_info(file)

        log.info(
            f"Added `{file.key} ({content_hash})`",
            uri=uri,
            dataset=self.name,
            from_store=store.uri,
            to_store=self._storage.uri,
        )
        return file

    def _put_file_info(self, file: OriginalFile) -> None:
        # store metadata
        self._storage.put(self._get_file_info_path(file.key), file, model=OriginalFile)
        # store hash reverse lookup
        self._storage.put(self._get_file_store_path(file.key, KEY), file.key.encode())

    def add_proxies(self, proxies: Iterable[CE]) -> None:
        path = self._get_entities_path()
        data = b"\n".join([orjson.dumps(p.to_full_dict()) for p in proxies])
        self._storage.put(path, data)

    def delete_file(self, key: str) -> None:
        self._storage.delete(self._get_file_info_path(key), ignore_errors=True)
        self._storage.delete(key, ignore_errors=True)



from typing import Annotated, Any, Optional, TypedDict

import orjson
import typer
from anystore.io import smart_open, smart_write
from pydantic import BaseModel
from rich.console import Console

from leakrfc import __version__
from leakrfc.archive import configure_archive, get_dataset
from leakrfc.archive.dataset import DatasetArchive
from leakrfc.crawl import crawl
from leakrfc.exceptions import ImproperlyConfigured
from leakrfc.export import export_dataset

# from leakrfc.ingest import ingest_dataset
from leakrfc.logging import configure_logging
from leakrfc.make import make_dataset
from leakrfc.settings import ArchiveSettings, Settings
from leakrfc.sync.aleph import sync_to_aleph
from leakrfc.sync.memorious import (
    get_file_name,
    get_file_name_strip_func,
    get_file_name_templ_func,
    import_memorious,
)

settings = Settings()
archive_settings = ArchiveSettings()
cli = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)
sync = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)
cli.add_typer(sync, name="sync")
console = Console(stderr=True)


class State(TypedDict):
    dataset: DatasetArchive | None


STATE: State = {"dataset": None}


def write_obj(obj: BaseModel | None, out: str) -> None:
    if out == "-":
        console.print(obj)
    else:
        if obj is not None:
            smart_write(
                out, orjson.dumps(obj.model_dump(), option=orjson.OPT_APPEND_NEWLINE)
            )


class ErrorHandler:
    def __enter__(self) -> Any:
        pass

    def __exit__(self, exc_cls, exc, _):
        if exc_cls is not None:
            if isinstance(exc, BrokenPipeError):
                return
            if settings.debug:
                raise exc
            console.print(f"[red][bold]{exc_cls.__name__}[/bold]: {exc}[/red]")
            raise typer.Exit(code=1)


class Dataset(ErrorHandler):
    def __enter__(self) -> DatasetArchive:
        if not STATE["dataset"]:
            e = ImproperlyConfigured("Specify dataset foreign_id with `-d` option!")
            if settings.debug:
                raise e
            console.print(f"[red][bold]{e.__class__.__name__}[/bold]: {e}[/red]")
            raise typer.Exit(code=1)
        return STATE["dataset"]


@cli.callback(invoke_without_command=True)
def cli_leakrfc(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
    dataset: Annotated[
        str | None, typer.Option("-d", help="Dataset foreign_id")
    ] = None,
):
    if version:
        console.print(__version__)
        raise typer.Exit()
    configure_logging()
    if dataset:
        STATE["dataset"] = get_dataset(dataset)


@cli.command("config")
def cli_config():
    """
    Print current runtime configuration for base archive or given dataset
    """
    with ErrorHandler():
        archive = configure_archive()
        dataset = STATE["dataset"]
        write_obj(settings, "-")
        write_obj(archive, "-")
        write_obj(archive_settings.cache, "-")
        write_obj(dataset, "-")


@cli.command("make")
def cli_make(
    out_uri: Annotated[str, typer.Option("-o")] = "-",
    use_cache: Annotated[Optional[bool], typer.Option(help="Use runtime cache")] = True,
    check_integrity: Annotated[
        Optional[bool], typer.Option(help="Check checksums")
    ] = True,
    cleanup: Annotated[
        Optional[bool], typer.Option(help="Cleanup (delete) unreferenced metadata")
    ] = True,
):
    """
    Make or update a leakrfc dataset and check integrity
    """
    with Dataset() as dataset:
        res = make_dataset(dataset, use_cache, check_integrity, cleanup)
        write_obj(res, out_uri)


@cli.command("get")
def cli_get(key: str, out_uri: Annotated[str, typer.Option("-o")] = "-"):
    """
    Retrieve a file from dataset archive and write to out uri (default: stdout)
    """
    with Dataset() as dataset:
        file = dataset.lookup_file(key)
        with dataset.open_file(file) as i, smart_open(out_uri, "wb") as o:
            o.write(i.read())


@cli.command("head")
def cli_head(key: str, out_uri: Annotated[str, typer.Option("-o")] = "-"):
    """
    Retrieve a file info from dataset archive and write to out uri (default: stdout)
    """
    with Dataset() as dataset:
        smart_write(
            out_uri,
            orjson.dumps(
                dataset.lookup_file(key).model_dump(),
                option=orjson.OPT_APPEND_NEWLINE,
            ),
        )


@cli.command("ls")
def cli_ls(
    out_uri: Annotated[str, typer.Option("-o")] = "-",
    keys: Annotated[bool, typer.Option(help="Show only keys")] = False,
):
    """
    List all files in dataset archive
    """
    with Dataset() as dataset:
        if keys:
            files = (f.key.encode() + b"\n" for f in dataset.iter_files())
        else:
            files = (
                orjson.dumps(f.model_dump(), option=orjson.OPT_APPEND_NEWLINE)
                for f in dataset.iter_files()
            )
        with smart_open(out_uri, "wb") as o:
            o.writelines(files)


@cli.command("crawl")
def cli_crawl(uri: str):
    """
    Crawl documents from local or remote sources
    """
    with Dataset() as dataset:
        crawl(uri, dataset)


@cli.command("export")
def cli_export(out: str):
    """
    Export a complete dataset in LeakRFC format
    """
    with Dataset() as dataset:
        write_obj(export_dataset(dataset, out), "-")


# @cli.command("ingest")
# def cli_ingest():
#     """
#     (Re-)ingest dataset
#     """
#     with Dataset() as dataset:
#         ingest_dataset(dataset)


@sync.command("memorious")
def cli_sync_memorious(
    uri: Annotated[str, typer.Option("-i")],
    use_cache: Annotated[Optional[bool], typer.Option(help="Use runtime cache")] = True,
    name_only: Annotated[
        Optional[bool], typer.Option(help="Use only file name as key")
    ] = False,
    strip_prefix: Annotated[
        Optional[str], typer.Option(help="Strip from file key prefix")
    ] = None,
    key_template: Annotated[
        Optional[str], typer.Option(help="Template to generate key")
    ] = None,
):
    """
    Sync a memorious data store into a leakrfc dataset
    """
    with Dataset() as dataset:
        if name_only:
            key_func = get_file_name
        elif strip_prefix:
            key_func = get_file_name_strip_func(strip_prefix)
        elif key_template:
            key_func = get_file_name_templ_func(key_template)
        else:
            key_func = None
        res = import_memorious(dataset, uri, key_func, use_cache=use_cache)
        write_obj(res, "-")


@sync.command("aleph")
def cli_sync_aleph(
    use_cache: Annotated[Optional[bool], typer.Option(help="Use runtime cache")] = True,
    host: Annotated[Optional[str], typer.Option(help="Aleph host")] = None,
    api_key: Annotated[Optional[str], typer.Option(help="Aleph api key")] = None,
    folder: Annotated[Optional[str], typer.Option(help="Base folder path")] = None,
    foreign_id: Annotated[
        Optional[str], typer.Option(help="Aleph foreign_id (if different from dataset)")
    ] = None,
):
    """
    Sync a leakrfc dataset to Aleph
    """
    with Dataset() as dataset:
        res = sync_to_aleph(
            dataset=dataset,
            host=host,
            api_key=api_key,
            prefix=folder,
            foreign_id=foreign_id,
            use_cache=use_cache,
        )
        write_obj(res, "-")



import mimetypes
from typing import Any

from anystore.io import smart_open
from anystore.types import Uri
from anystore.util import make_checksum as _make_checksum
from jinja2 import Template
from pantomime import DEFAULT, normalize_mimetype


def make_checksum(uri: Uri, algorithm: str | None = "sha1") -> str:
    with smart_open(str(uri)) as io:
        return _make_checksum(io, algorithm)


def make_ch_key(ch: str) -> str:
    if len(ch) < 6:
        raise ValueError(f"Invalid checksum: `{ch}`")
    return "/".join((ch[:2], ch[2:4], ch[4:6], ch))


def guess_mimetype(value: Any) -> str | None:
    if not value:
        return
    guess = normalize_mimetype(value)
    if guess != DEFAULT:
        return guess
    mtype, _ = mimetypes.guess_type(value)
    return normalize_mimetype(mtype)


def render(tmpl: str, data: dict[str, Any]) -> str:
    template = Template(tmpl)
    return template.render(**data)



from anystore.store import ZipStore
from anystore.types import StrGenerator, Uri
from anystore.worker import WorkerStatus
from fsspec.implementations.zip import zipfile

from leakrfc.archive import DatasetArchive
from leakrfc.logging import get_logger
from leakrfc.worker import DatasetWorker

log = get_logger(__name__)


class ExportWorker(DatasetWorker):
    def __init__(self, out: ZipStore, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.out = out
        self.consumer_threads = 1

    def get_tasks(self) -> StrGenerator:
        yield from self.dataset._storage.iterate_keys()

    def handle_task(self, task: str) -> None:
        self.log_info(f"Adding file `{task}` ...")
        with self.dataset._storage.open(task) as i:
            with self.out.open(f"{self.dataset.name}/{task}", mode="wb") as o:
                o.write(i.read())

    def done(self) -> None:
        self.log_info(f"Exporting dataset `{self.dataset.name}`: Done.")


def export_dataset(dataset: DatasetArchive, uri: Uri) -> WorkerStatus:
    out = ZipStore(uri=uri, backend_config={"compression": zipfile.ZIP_LZMA})
    worker = ExportWorker(out, dataset)
    worker.log_info(f"Exporting dataset `{dataset.name}` ...")
    return worker.run()



from fastapi import Depends, FastAPI, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import StreamingResponse

from leakrfc import __version__
from leakrfc.api.auth import (
    Token,
    create_access_token,
    ensure_auth_context,
    ensure_token_context,
)
from leakrfc.api.util import Context, Errors, ensure_path_context, stream_file
from leakrfc.archive import get_archive
from leakrfc.logging import get_logger
from leakrfc.settings import ApiSettings, Settings

log = get_logger(__name__)

settings = Settings()
api_settings = ApiSettings()

app = FastAPI(
    debug=settings.debug,
    title=api_settings.title,
    # contact=api_settings.contact,
    description=api_settings.description,
    redoc_url="/",
    version=__version__,
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=[str(api_settings.allowed_origin)],
    allow_methods=["HEAD", "OPTIONS", "GET"],
)

archive = get_archive()
log.info(f"Archive: `{archive._storage.uri}`")


if settings.debug:

    @app.get("/{dataset}/{key}/token")
    async def get_token(
        response: Response,
        ctx: Context = Depends(ensure_path_context),
        exp: int | None = None,
    ) -> Token:
        """
        Obtain a temporary access token. This is for development purposes only!
        """
        with Errors():
            response.headers.update(ctx.headers)
            return Token(
                access_token=create_access_token(ctx.dataset, ctx.key, exp),
                token_type="Bearer",
            )


@app.head("/file")
async def head_file_by_token(
    response: Response, ctx: Context = Depends(ensure_auth_context)
) -> None:
    """
    Get metadata of a private file
    """
    with Errors():
        response.headers.update(ctx.headers)


@app.get("/file", response_model=None)
async def get_file_by_token(
    ctx: Context = Depends(ensure_auth_context),
) -> StreamingResponse:
    """
    Stream contents of a private file
    """
    with Errors():
        return stream_file(ctx)


@app.head("/{dataset}/{key}")
async def head_file(
    response: Response, ctx: Context = Depends(ensure_path_context)
) -> None:
    """
    Get metadata of a public file
    """
    with Errors():
        response.headers.update(ctx.headers)


@app.get("/{dataset}/{key}", response_model=None)
async def get_file(ctx: Context = Depends(ensure_path_context)) -> StreamingResponse:
    """
    Stream contents of a public file
    """
    with Errors():
        return stream_file(ctx)


@app.get("/api/2/archive", response_model=None)
async def legacy_aleph_api(
    ctx: Context = Depends(ensure_token_context),
) -> StreamingResponse:
    """
    Stream contents of a file, mimic Aleph servicelayer api to act as a drop-in
    replacement
    """
    with Errors():
        return stream_file(ctx)



from typing import Any

from anystore.worker import Worker

from leakrfc.archive.dataset import DatasetArchive
from leakrfc.logging import get_logger
from leakrfc.settings import Settings

log = get_logger(__name__)

settings = Settings()


class DatasetWorker(Worker):
    def __init__(
        self, dataset: DatasetArchive, use_cache: bool | None = True, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.dataset = dataset
        self.use_cache = use_cache

    def get_tasks(self) -> Any:
        yield from self.dataset.iter_files()

    def log_info(self, msg: str, **ctx) -> None:
        ctx = {
            "dataset": self.dataset.name,
            "storage": self.dataset._storage.uri,
            **ctx,
        }
        log.info(msg, **ctx)

    def log_warning(self, msg: str, **ctx) -> None:
        ctx = {
            "dataset": self.dataset.name,
            "storage": self.dataset._storage.uri,
            **ctx,
        }
        log.warning(msg, **ctx)

    def log_error(self, msg: str, **ctx) -> None:
        ctx = {
            "dataset": self.dataset.name,
            "storage": self.dataset._storage.uri,
            **ctx,
        }
        log.error(msg, **ctx)

    def exception(self, task: Any, e: Exception) -> None:
        self.log_error(
            f"Error while handling task: {e.__class__.__name__}: {e}",
            task=task,
        )
        if settings.debug:
            raise e



from anystore import anycache
from anystore.store.fs import DoesNotExist
from anystore.util import clean_dict
from fastapi import HTTPException
from fastapi.responses import StreamingResponse
from pydantic import BaseModel

from leakrfc import __version__
from leakrfc.archive import archive
from leakrfc.archive.cache import get_cache
from leakrfc.logging import get_logger
from leakrfc.model import File
from leakrfc.settings import Settings

settings = Settings()
log = get_logger(__name__)
DEFAULT_ERROR = HTTPException(404)
BASE_HEADER = {"x-leakrfc-version": __version__}


def get_base_header(dataset: str, key: str | None = None) -> dict[str, str]:
    return clean_dict(
        {**BASE_HEADER, "x-leakrfc-dataset": dataset, "x-leakrfc-key": key}
    )


def get_file_header(file: File) -> dict[str, str]:
    return clean_dict(
        {
            **get_base_header(file.dataset, file.content_hash),
            "x-leakrfc-file": file.name,
            "x-leakrfc-size": str(file.size),
            "x-mimetype": file.mimetype,
            "content-type": file.mimetype,
        }
    )


class Context(BaseModel):
    dataset: str
    key: str
    file: File

    @property
    def headers(self) -> dict[str, str]:
        return get_file_header(self.file)


class Errors:
    def __enter__(self):
        pass

    def __exit__(self, exc_cls, exc, _):
        if exc_cls is not None:
            log.error(f"{exc_cls.__name__}: `{exc}`")
            if not settings.debug:
                # always just 404 for information hiding
                raise DEFAULT_ERROR
            else:
                if exc_cls == DoesNotExist:
                    raise DEFAULT_ERROR
                raise exc


@anycache(store=get_cache(), key_func=lambda d, k: f"api/file/{d}/{k}", model=File)
def get_file_info(dataset: str, key: str) -> File | None:
    storage = archive.get_dataset(dataset)
    return storage.lookup_file_by_hash(key)


def ensure_path_context(dataset: str, key: str) -> Context:
    with Errors():
        return Context(dataset=dataset, key=key, file=get_file_info(dataset, key))


def stream_file(ctx: Context) -> StreamingResponse:
    storage = archive.get_dataset(ctx.dataset)
    file = storage.lookup_file_by_hash(ctx.key)
    return StreamingResponse(
        storage.stream_file(file),
        headers=ctx.headers,
        media_type=ctx.file.mimetype,
    )



"""
Sync Aleph collections into leakrfc or vice versa via `alephclient`
"""

from datetime import datetime
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from anystore import anycache
from anystore.store.virtual import get_virtual
from anystore.worker import WorkerStatus

from leakrfc.archive.cache import get_cache
from leakrfc.archive.dataset import DatasetArchive
from leakrfc.connectors import aleph
from leakrfc.model import OriginalFile
from leakrfc.worker import DatasetWorker


def _make_cache_key(self: "AlephUploadWorker", *parts: str) -> str:
    host = urlparse(self.host).netloc
    base = f"aleph/upload/{host}/{self.dataset.name}/"
    return base + "/".join(parts)


def get_upload_cache_key(self: "AlephUploadWorker", file: OriginalFile) -> str | None:
    if self.use_cache:
        return _make_cache_key(self, file.key)


def get_parent_cache_key(
    self: "AlephUploadWorker", key: str, prefix: str | None = None
) -> str | None:
    if self.use_cache:
        parts = [str(Path(key).parent)]
        if prefix:
            parts += prefix
        return _make_cache_key(self, *parts)


class AlephUploadStatus(WorkerStatus):
    added: int = 0
    folders_created: int = 0


class AlephUploadWorker(DatasetWorker):
    """
    Sync leakrfc dataset to an Aleph instance
    """

    def __init__(
        self,
        host: str | None = None,
        api_key: str | None = None,
        prefix: str | None = None,
        foreign_id: str | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tmp = get_virtual(f"leakrfc-{self.dataset.name}-")
        self.api = aleph.get_api(host, api_key)
        self.host = aleph.get_host(self.api)
        self.foreign_id = foreign_id or self.dataset.name
        self.collection_id = aleph.get_or_create_collection_id(
            self.foreign_id, self.api
        )
        self.prefix = prefix
        self.consumer_threads = min(10, self.consumer_threads)  # urllib connection pool

    @anycache(store=get_cache(), key_func=get_parent_cache_key)
    def get_parent(self, key: str, prefix: str | None = None) -> dict[str, str] | None:
        with self.lock:
            p = Path(key)
            if prefix:
                p = prefix / p
            parent_path = str(p.parent)
            if not parent_path or parent_path == ".":
                return
            parent = {"id": aleph.make_folders(parent_path, self.collection_id)}
            self.count(folders_created=1)
            return parent

    @anycache(store=get_cache(), key_func=get_upload_cache_key)
    def handle_task(self, task: OriginalFile) -> dict[str, Any]:
        res = {
            "uploaded_at": datetime.now().isoformat(),
            "dataset": self.dataset.name,
            "host": self.host,
        }
        self.log_info(
            f"Uploading `{task.key}` ({task.content_hash}) ...",
            aleph=self.host,
            foreign_id=self.foreign_id,
        )
        metadata = {**task.extra, "file_name": task.name, "foreign_id": task.key}
        metadata["source_url"] = metadata.get("url")
        parent = self.get_parent(task.key, self.prefix)
        if parent:
            metadata["parent"] = parent
        tmp_key = self.tmp.download(
            self.dataset._make_path(task.key), self.dataset._storage
        )
        tmp_path = urlparse(self.tmp.store.get_key(tmp_key)).path
        res.update(
            self.api.ingest_upload(
                self.collection_id, Path(tmp_path), metadata=metadata
            )
        )
        self.tmp.cleanup(tmp_key)
        self.log_info(
            f"Upload complete. Aleph id: `{res['id']}`",
            content_hash=task.content_hash,
            aleph=self.host,
            file=task.key,
            foreign_id=self.foreign_id,
        )
        self.count(uploaded=1)
        return res

    def done(self) -> None:
        self.tmp.cleanup()
        self.log_info("Syncing to Aleph: Done")


def sync_to_aleph(
    dataset: DatasetArchive,
    host: str | None,
    api_key: str | None,
    prefix: str | None = None,
    foreign_id: str | None = None,
    use_cache: bool | None = True,
) -> AlephUploadStatus:
    worker = AlephUploadWorker(
        dataset=dataset,
        host=host,
        api_key=api_key,
        prefix=prefix,
        foreign_id=foreign_id,
        use_cache=use_cache,
    )
    worker.log_info(f"Starting sync to Aleph `{worker.host}` ...")
    return worker.run()



"""
https://fastapi.tiangolo.com/tutorial/security/oauth2-jwt/

Authorization expects an encrypted bearer token with the dataset and key lookup
in the subject ({"sub": "<dataset>/<key>"}). Therefore, clients need to be able
to create such tokens (knowing the secret key) and handle dataset permissions.

Tokens should have a short expiration (via `exp` property in payload).
"""

from datetime import UTC, datetime, timedelta

import jwt
from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from pydantic import BaseModel

from leakrfc.api.util import DEFAULT_ERROR, Context, ensure_path_context
from leakrfc.logging import get_logger
from leakrfc.settings import ApiSettings

settings = ApiSettings()
log = get_logger(__name__)
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/", auto_error=False)


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    dataset: str
    key: str

    def __init__(self, **data):
        dataset, key = data.pop("sub").split("/", 1)
        data["dataset"] = dataset
        data["key"] = key
        super().__init__(**data)


def create_access_token(dataset: str, key: str, exp: int | None = None) -> str:
    exp = datetime.now(UTC) + timedelta(minutes=exp or settings.access_token_expire)
    data = {"sub": f"{dataset}/{key}", "exp": exp}
    return jwt.encode(
        data, settings.secret_key, algorithm=settings.access_token_algorithm
    )


def ensure_token_context(token: str) -> Context:
    """Get context from url query argument"""

    if not token:
        log.error("Auth: no token")
        raise DEFAULT_ERROR
    try:
        payload = jwt.decode(
            token,
            settings.secret_key,
            algorithms=[settings.access_token_algorithm],
            verify=True,
        )
        data = TokenData(**payload)
        return ensure_path_context(data.dataset, data.key)
    except Exception as e:
        log.error(f"Invalid token: `{e}`", token=token)
        raise DEFAULT_ERROR


def ensure_auth_context(token: str = Depends(oauth2_scheme)) -> Context:
    """Get context from Authorization header"""

    return ensure_token_context(token)


# def ensure_aleph_context(token: str) -> Context:
#     """Decode legacy Aleph archive token"""

#     if not token:
#         raise DEFAULT_ERROR
#     try:
#         payload = jwt.decode(
#             token, settings.aleph_secret_key, algorithms=["HS256"], verify=True
#         )
#         key = payload["c"]
#         return ensure_path_context(data.dataset, key)
#     except Exception as e:
#         log.error(f"Invalid Aleph token: `{e}`", token=token)
#         raise DEFAULT_ERROR
